"""
Tue Herlau, 2022: I am bundling a version of the pyminifier library since the version maintained by the original author has
not been updated for a while and will no longer install via pypi on the python versions I want to target.

All code in this folder is either a copy or very small modification of code with this copyright statement:

#
#       Copyright 2013 Liftoff Software Corporation
#
# For license information see LICENSE.txt
The licence file can be found on the original repostiory.

All rights belong to the original author.
"""